#!/bin/bash
if [ $# -lt 1 ]; then
  echo "Install MACSIMUS. Call by:"
  echo "  ./install.sh COMPILER"
  echo "or"
  echo "  bash install.sh COMPILER"
  echo "Where COMPILER is the C compiler name:"
  echo "  gcc = GCC compiler - recommended"
  echo "        the default optimization level is -O2 -ffast-math"
  echo "        -O3 may work, but has not been extensively tested"
  echo "  colorgcc = as above, colorized messages"
  echo "  icc = Intel compiler - faster code than gcc, less reliable on AMD"
  echo "        the default optimization level is -O2"
  echo "        flag -fast is not reliable (Trace/breakpoint trap)"
  echo "        locale must be set, LANG=C is recommended"
  echo "        (bash: use export LANG=C, tcsh: use setenv LANG C)"
  echo "  pgcc = Portland Group, Inc. (PGI) compiler"
  echo "        BUG: RANDOM NUMBERS DO NOT WORK (to be solved later)"
  echo "  cc = \"any\" compiler, edit compile.mmk and check your system for options"
  echo "For implementing a new compiler, see compile.mmk"
  if [ -n "$CC" ] ; then
    echo "Example (based on your \$CC=$CC):"
    echo "  ./install.sh $CC"
  else
    echo "Example:"
    echo "  ./install.sh gcc"
  fi
  exit
fi

# makemake option: now unix and linux are equivalent
UNIX=linux
CC=$1

# no spaces in the path allowed
if [ "${PWD/ /z}" != "${PWD}" ] ; then
  echo "Sorry, I cannot handle spaces in PWD=\"$PWD\"!"
  echo "Please install MACSIMUS in a path not containing spaces"
  exit
fi

# check if the selected compiler exists
if ! which $CC ; then
  echo "compiler $1 not found"
  exit
fi

# check bin/makeone.sh
if [ ! -e bin/makeone.sh ]; then
  echo "ERROR bin/makeone.sh not found (MACSIMUS incorrectly unzipped?)"
  exit
fi
chmod +x bin/makeone.sh

# using lib64 if exists (probably no longer needed)
# if [ -d /usr/X11R6/lib64 ] ; then
# cat <<EOF | ed graph.mmk
# ,s:/usr/X11R6/lib:/usr/X11R6/lib64:g
# w
# q
# EOF
# echo /usr/X11R6/lib64 used
# fi

# environment for cook/configure.sh
cat > makemake.sh <<EOF
export CC=$CC
# this file is used by script cook/configure.sh
EOF

# export MACSIMUS home directory for makemake
cat > home.mmk <<EOF
# this file defines the MACSIMUS home directory
# it is !included by all metamake files
# non-! statements will become a part of the generated makefile files

EOF
echo '!home='${PWD} >> home.mmk
echo "home.mmk generated"

echo
echo ">>> int4 check"
cd gen/
$1 -o int4 int4.c 
if ! ./int4 ; then
  echo "This is file $PWD/int4.h:"
  echo "----------------------------------------------------"
  cat $PWD/int4.h
  echo "----------------------------------------------------"

  echo "Continue anyway (y/N)?"
  read
  if [ "$REPLY" != "y" ] ; then
    exit
  fi
fi
cd ..

echo
echo ">>> compiling the makemake utility"
cd c/
rm *.o
$1 -o makemake makemake.c
if ! ./makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
rm makemake # will be recompiled

echo
echo ">>> compiling general utilities from subdirectory c/"
for f in \
  makemake \
  tabproc smoothg extract \
  mergetab tab endian start sortcite filttab lemon eqfield sum \
  lattice sumetc blocktab eddata ice histogr runsum exp2pow \
  runint binrepl liat lc
do
  ../bin/makeone.sh $f
done
  echo $CC -o ev -O2 -Wall -DCALC=0 ev.c -I../gen -lm -lncurses
  $CC -o ev -O2 -Wall -DCALC=0 ev.c -I../gen -lm -lncurses
  echo $CC -o evu -O2 -Wall -DCALC=1 ev.c -I../gen -lm -lncurses
  $CC -o evu -O2 -Wall -DCALC=1 ev.c -I../gen -lm -lncurses
  mv ev ev.exe evu evu.exe ../bin 2> /dev/null
cd ..

# support of fitting by plot
cd c/fit4plot
rm *.o *.ol *.ox
../../bin/makemake $UNIX $1 $2 long
mv makefile makefile.long
../../bin/makemake $UNIX $1 $2 high
mv makefile makefile.high
../../bin/makemake $UNIX $1 $2
mv makefile makefile.double
cd ../..

echo
echo ">>> compiling blend (lj version), pdb, and similar from blend/"
cd blend/
if ! ../bin/makemake $UNIX lj $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi

rm *.o
for f in \
  blend pdb showpro makepept blefilt ramachan pdb2pdb bonds
do
  ../bin/makeone.sh $f
done

echo
echo ">>> installing bin versions of common force fields from blend/data/"
cd data
rm gromos96.bin charmm19.bin charmm21.bin charmm22.bin
export BLENDPATH=.
../../bin/blend -b gromos96.par
../../bin/blend -b charmm19.par
../../bin/blend -b charmm21.par
../../bin/blend -b charmm22.par
cd ../..

# compiling graphical utilities
cd show/
if ! ../bin/makemake $UNIX $1 $2; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
rm *.o
for f in \
  show plot stereo tomoil plbmatch
do
  ../bin/makeone.sh $f
done
cd ..

echo
echo ">>> compiling different simulation utilities from util/"
cd util/
if ! ../bin/makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
rm *.o
for f in \
  autocorr cppak molcfg plb2diff plbmsd plbpak rdfg sfourier showcp \
  spectrum staprt \
  coordn cp2cp cutprt plbcut plbcenter hbonds density densprof plbfilt \
  plbmerge plbinfo plb2nbr plb2cryst shownear plbstack plbreplicate plbsites \
  plb2asc plbcut plb2plb plbconv asc2plb cfg2atm \
  plb2cfg cfg2plb cfg2atm plbframe plbscale plbbox plbtran mol2mol
do
  ../bin/makeone.sh $f
done
cd ..

echo
echo ">>> compiling the raytracer + related software from ray/"
echo "(more utilities available upon request)"
cd ray/
if ! ../bin/makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
rm *.o
for f in \
  ray ppm2ps ppminfo ppmscale ppm2ppm
do
  ../bin/makeone.sh $f
done
cd ..

echo
echo ">>> compiling three cook versions, see MACSIMUS/cook/exe/"
cd cook/exe/free
if ! ../../../bin/makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
../../../bin/makeone.sh cookfree
mv cookfree ../../../bin/
cd -

cd cook/exe/cutelst
if ! ../../../bin/makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
../../../bin/makeone.sh cookcec
mv cookcec ../../../bin/
cd -

cd cook/exe/ewaldslab
if ! ../../../bin/makemake $UNIX $1 $2 ; then
  echo "makemake problem (makemake not found, bad metamake or !include *.mmk)"
  exit
fi
../../../bin/makeone.sh cookewslc
mv cookewslc ../../../bin/cook
cd -

echo
cd bin
# LC_COLLATE=C added by T.Trnka (../bin.lst is of C locale order)
LC_COLLATE=C \ls -1 > ../instbin.lst
# this is because of CygWin
cat <<EOF | ed ../instbin.lst
,s:\.exe::g
w
q
EOF
echo "Comparing (using diff):"
echo "  the contents of ${PWD}/ (instbin.lst)"
echo "  with the expected listing bin.lst"
echo "----------------------------------------------------------------------------"
if diff ../bin.lst ../instbin.lst ; then
  echo "      NO DIFFERENCE DETECTED - MACSIMUS HAS BEEN SUCCESSFULLY COMPILED"
else
  echo "----------------------------------------------------------------------------"
  echo " ERROR: COMPONENTS LISTED ABOVE WITH < HAVE NOT BEEN SUCCESSFULLY COMPILED"
  echo " (or the following utilities are missing: diff ed)."
fi
echo "----------------------------------------------------------------------------"

cd ..

echo
echo "If you wish to use the MACSIMUS calculators ev (numbers) and evu (+units) or"
echo "  extension-based start, you must install the definition files."
echo "Hint: get help by running ev|evu|start without parameters"
echo "Copy .evdata,.evudata,.startdata to your home (y/N)?"
read
if [ "$REPLY" == "y" ] ; then
  cp -ip .evdata .evudata .startdata ~/
fi

if [ -d ~/.config/mc ] ; then
  echo
  echo "INFO: ~/.config/mc found, it looks you are using Midnight Commander"
  if [ -e ~/.config/mc/mc.ext ] ; then
    echo "INFO: extension association file ~/.config/mc/mc.ext found!"
    echo "To activate the MACSIMUS extensions (incl. common graphic files via"
    echo "display/ImageMagick), you should install macsimus/mc.ext:"
    echo "  a = Append macsimus/mc.ext to ~/.config/mc/mc.ext"
    echo "  r = Replace ~/.config/mc/mc.ext by macsimus/mc.ext (with backup~)"
    echo "> n = Nothing (default)"
  else
    echo "INFO: no extension association file ~/.config/mc/mc.ext found!"
    echo "To activate the MACSIMUS extensions (incl. common graphic files via"
    echo "display/ImageMagick), you should install macsimus/mc.ext:"
    echo "  c = Copy macsimus/mc.ext to ~/.config/mc/mc.ext"
    echo "> n = Nothing (default)"
  fi

  read
  case $REPLY in 
    a | A ) cat mc.ext >>  ~/.config/mc/mc.ext ;;
    c | C | r | R )
      [ -e  ~/.config/mc/mc.ext ] && mv  ~/.config/mc/mc.ext ~/.config/mc/mc.ext~
      cp mc.ext ~/.config/mc/mc.ext
      ;;
    * ) ;;
  esac
fi

if [ -d ~/.mc ] ; then
  echo
  echo "INFO: ~/.mc found, it looks you are using an OBSOLETE Midnight Commander"
  if [ -e ~/.mc/bindings ] ; then
    echo "INFO: extension association file ~/.mc/bindings found!"
    echo "To activate the MACSIMUS extensions (incl. common graphic files via"
    echo "display/ImageMagick), you should install macsimus/mc.ext:"
    echo "  a = Append macsimus/mc.ext to ~/.mc/bindings"
    echo "  r = Replace ~/.mc/bindings by macsimus/mc.ext (with backup~)"
    echo "> n = Nothing (default)"
  else
    echo "INFO: no extension association file ~/.mc/bindings found!"
    echo "To activate the MACSIMUS extensions (incl. common graphic files via"
    echo "display/ImageMagick), you should install macsimus/mc.ext:"
    echo "  c = Copy macsimus/mc.ext to ~/.mc/bindings"
    echo "> n = Nothing (default)"
  fi

  read
  case $REPLY in 
    a | A ) cat mc.ext >>  ~/.mc/bindings ;;
    c | C | r | R )
      [ -e  ~/.mc/bindings ] && mv ~/.mc/bindings ~/.mc/bindings~
      cp mc.ext ~/.mc/bindings
      ;;
    * ) ;;
  esac
fi

echo =============================================================================
echo "To finish installation:"
echo "* IF you are using sh or bash,"
echo "  THEN append the following lines to your ~/.bashrc file:"
echo "    export BLENDPATH=${PWD}/blend/data"
echo "    export PATH=${PWD}/bin:\$PATH"
echo "* ELSE use the syntax of your shell."
echo
echo "* Three Lennard-Jones versions of cook has been precompiled:"
echo "  - cook = Ewald sums, slab + density profile support, linked-cell list"
echo "  - cookcec = cutoff electrostatics, simplest all-neighbor algorithm"
echo "  - cookfree = in vacuum"
echo "* You may customize and compile your own cook* version(s) as follows:"
echo "  - go to directory" ${PWD}/cook
echo "    and run script ./configure.sh"
echo "  - alternatively, customize" ${PWD}/generic/metamake
echo "    and" ${PWD}/generic/simopt.h "by hand"
echo "    and then run makemake $1 and make cook*"
echo
echo "* The basic Lennard-Jones version of blend has been precompiled."
echo "  To compile other blend versions:"
echo "  - follow the instructions in ${PWD}/blend/metamake"
echo
echo "Most utilities print help if run without any option, sometimes -h is needed"
echo "If a utility mentioned after \"See also:\" is not present, please let me know"
echo =============================================================================

echo
X=later
echo "Would you like to start a simple test of MACSIMUS now (y/N)?"
read
if [ "$REPLY" == "y" ] ; then
  export BLENDPATH=${PWD}/blend/data
  export PATH=${PWD}/bin:$PATH
  cd examples
  echo "NB: the environment has been set within the script."
  ./test.sh
  X=again
  cd ..
fi

echo
echo "You can run the test $X as ./test.sh from ${PWD}/examples/"
echo
echo "MORE: To show normal mode vibrations of selected molecules:"
cd blend/che
echo "  cd $PWD"
echo "Then EITHER from the command prompt as: "
echo "  CHE=2 che.sh MOLECULE.che"
echo "OR using the Midnight Commander (if the MACSIMUS associations are installed):"
echo "  CHE=2 mc"
echo "and double click a molecule."
echo "If minimized, press button [finish] or hotkey ."
